#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <vector>

#include <agile.h>

#if !defined GL_GLEXT_PROTOTYPES
	#define GL_GLEXT_PROTOTYPES
#endif

#include <GLES2/gl2.h>
#include <GLES2/gl2ext.h>
#define YS_GL_ES2
typedef double GLdouble;

#include <ysgl.h>
#include <uwpdirlist.h>

#include <EGL/egl.h>
#include <EGL/eglext.h>
#include <EGL/eglplatform.h>
#include <angle_windowsstore.h>



using namespace Windows::ApplicationModel::Core;
using namespace Windows::ApplicationModel::Activation;
using namespace Windows::UI::Core;
using namespace Windows::UI::Input;
using namespace Windows::Foundation;
using namespace Windows::Foundation::Collections;
using namespace Windows::Graphics::Display;
using namespace Windows::System;

using namespace Platform;



ref class SingleWindowAppView sealed : public Windows::ApplicationModel::Core::IFrameworkView
{
public:
	typedef SingleWindowAppView THISCLASS;

private:  // Cannot have a public member variable?
	bool isWindowClosed,isWindowVisible;
	Platform::Agile<CoreWindow> hWnd;

	static THISCLASS ^CurrentView;

	EGLDisplay mEglDisplay;
	EGLContext mEglContext;
	EGLSurface mEglSurface;

	bool first;
	YsUwpFileList fList;

	static std::vector <unsigned char> RenderTextToBitmap(int &w,int &h,const wchar_t wMsg[]);

public:
	static THISCLASS ^GetCurrentView(void);

	SingleWindowAppView();

	virtual void Initialize(Windows::ApplicationModel::Core::CoreApplicationView^ applicationView);
	virtual void SetWindow(Windows::UI::Core::CoreWindow^ window);
	virtual void Load(Platform::String^ entryPoint);
	virtual void Run();
	virtual void Uninitialize();

	virtual void OnActivated(CoreApplicationView^ applicationView, IActivatedEventArgs^ args);
	virtual void OnVisibilityChanged(CoreWindow^ sender, VisibilityChangedEventArgs^ args);
	virtual void OnWindowClosed(CoreWindow^ sender, CoreWindowEventArgs^ args);

private:
	void InitializeOpenGLES2(void);
	EGLBoolean InitializeEGLDisplayDefault(void);
	EGLBoolean InitializeEGLDisplay(int level);
	std::vector <EGLint> MakeDisplayAttribute(int level) const;
	void SetUpConfigAttrib(EGLint configAttrib[],int depthBit);
	void CleanUpOpenGLES2(void);
};

/* static */ SingleWindowAppView ^SingleWindowAppView::CurrentView;

/* static */ SingleWindowAppView ^SingleWindowAppView::GetCurrentView(void)
{
	return CurrentView;
}

/* static */ std::vector <unsigned char> SingleWindowAppView::RenderTextToBitmap(int &w,int &h,const wchar_t wMsg[])
{
	const int bmpWid=640,bmpHei=120;

	auto device=Microsoft::Graphics::Canvas::CanvasDevice::GetSharedDevice();
	auto bitmap=ref new Microsoft::Graphics::Canvas::CanvasRenderTarget(device,(float)bmpWid,(float)bmpHei,96);

	auto ds=bitmap->CreateDrawingSession();
	ds->Clear(Windows::UI::Colors::Green);

	auto txtFmt=ref new Microsoft::Graphics::Canvas::Text::CanvasTextFormat();
	txtFmt->FontSize=bitmap->ConvertPixelsToDips(20);

	auto wStr=ref new Platform::String(wMsg);

	auto txtLayout=ref new Microsoft::Graphics::Canvas::Text::CanvasTextLayout(ds,wStr,txtFmt,(float)bmpWid,(float)bmpHei);
	auto rounding=Microsoft::Graphics::Canvas::CanvasDpiRounding::Floor;
	int renderX=(int)bitmap->ConvertDipsToPixels(txtLayout->LayoutBounds.X,rounding);
	int renderY=(int)bitmap->ConvertDipsToPixels(txtLayout->LayoutBounds.Y,rounding);
	int renderW=(int)bitmap->ConvertDipsToPixels(txtLayout->LayoutBounds.Width,rounding);
	int renderH=(int)bitmap->ConvertDipsToPixels(txtLayout->LayoutBounds.Height,rounding);

	ds->DrawTextLayout(txtLayout,0,0,Windows::UI::Colors::Orange);

	ds->Flush(); // <- Don't forget this, or you get nothing.

	// When you need the bitmap size of the text on the bitmap,
	// you can do it by CanvasTextLayout.
	// http://stackoverflow.com/questions/30696838/how-to-calculate-the-size-of-a-piece-of-text-in-win2d

	// Windows bitmap starts at top-left, while OpenGL bitmap at bottom-left.
	// Bitmap needs to be flipped.
	w=renderW;
	h=renderH;

	auto rawBitmap=bitmap->GetPixelBytes();
	std::vector <unsigned char> reallyRawBitmap;
	reallyRawBitmap.resize(renderW*renderH*4);
	{
		int x=0,y=0;
		for(int y=0; y<renderH; ++y)
		{
			for(int x=0; x<renderW; ++x)
			{
				if(x<bmpWid && y<bmpHei)
				{
					// Windows -> BGRA   OpenGL -> RGBA  Need conversion.
					reallyRawBitmap[((h-y-1)*w+x)*4  ]=rawBitmap[(y*bmpWid+x)*4+2];
					reallyRawBitmap[((h-y-1)*w+x)*4+1]=rawBitmap[(y*bmpWid+x)*4+1];
					reallyRawBitmap[((h-y-1)*w+x)*4+2]=rawBitmap[(y*bmpWid+x)*4];
					reallyRawBitmap[((h-y-1)*w+x)*4+3]=rawBitmap[(y*bmpWid+x)*4+3];
				}
				else
				{
					reallyRawBitmap[((h-y-1)*w+x)*4  ]=0;
					reallyRawBitmap[((h-y-1)*w+x)*4+1]=0;
					reallyRawBitmap[((h-y-1)*w+x)*4+2]=0;
					reallyRawBitmap[((h-y-1)*w+x)*4+3]=0;
				}
			}
		}
	}

	return reallyRawBitmap;
}

SingleWindowAppView::SingleWindowAppView()
{
	first=true;
	isWindowClosed=false;
	isWindowVisible=false;

	mEglDisplay=EGL_NO_DISPLAY;
	mEglContext=EGL_NO_CONTEXT;
	mEglSurface=EGL_NO_SURFACE;

	CurrentView=this;
}

void SingleWindowAppView::Initialize(Windows::ApplicationModel::Core::CoreApplicationView^ applicationView)
{
	applicationView->Activated+=
		ref new TypedEventHandler<CoreApplicationView^, IActivatedEventArgs^>(this,&THISCLASS::OnActivated);
}
void SingleWindowAppView::SetWindow(Windows::UI::Core::CoreWindow^ window)
{
	hWnd=window;

	window->VisibilityChanged+=
		ref new TypedEventHandler<CoreWindow^,VisibilityChangedEventArgs^>(this,&THISCLASS::OnVisibilityChanged);
	window->Closed+=
		ref new TypedEventHandler<CoreWindow^,CoreWindowEventArgs^>(this,&THISCLASS::OnWindowClosed);

	InitializeOpenGLES2(); // Example initializes OpenGL ES context in here.
}
void SingleWindowAppView::Load(Platform::String^ entryPoint)
{
	YsGLSLCreateSharedRenderer();
}
void SingleWindowAppView::Run()
{
	while(!isWindowClosed)
	{
		if (isWindowVisible)
		{
			if(true==first)
			{
				auto package=Windows::ApplicationModel::Package::Current;
				auto instDir=package->InstalledLocation->Path->Data();

				fList.MakeFileList(instDir);
				first=false;
			}

			// This doesn't seem to wait for an event.  Like a PeekMessage
			CoreWindow::GetForCurrentThread()->Dispatcher->ProcessEvents(CoreProcessEventsOption::ProcessAllIfPresent);

			glClearColor(0,0,1,0);
			glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);

			EGLint w=0,h=0;
			eglQuerySurface(mEglDisplay, mEglSurface, EGL_WIDTH, &w);
			eglQuerySurface(mEglDisplay, mEglSurface, EGL_HEIGHT, &h);
			if(0<w && 0<h)
			{
				int y=0;
				glViewport(0,0,w,h);
				for(auto &f : fList.fileList)
				{
					int bmpW,bmpH;
					auto bmp=RenderTextToBitmap(bmpW,bmpH,f.nm.data());

					// In the BitmapRenderer, it just create and render a texture.
					// I don't explain OpenGL ES itself here.
					auto renderer=YsGLSLSharedBitmapRenderer();
					YsGLSLUseBitmapRenderer(renderer);
					YsGLSLRenderRGBABitmap2D(renderer,100,100+y,YSGLSL_HALIGN_LEFT,YSGLSL_VALIGN_TOP,bmpW,bmpH,bmp.data());
					YsGLSLEndUseBitmapRenderer(renderer);
					y+=bmpH;
				}
			}

			if (eglSwapBuffers(mEglDisplay, mEglSurface) != GL_TRUE)
			{
				YsGLSLDeleteSharedRenderer();
				CleanUpOpenGLES2();
				InitializeOpenGLES2();
				YsGLSLCreateSharedRenderer();
			}
		}
		else
		{
			// This seems to wait until an event is thrown.
			CoreWindow::GetForCurrentThread()->Dispatcher->ProcessEvents(CoreProcessEventsOption::ProcessOneAndAllPending);
		}
	}

	CleanUpOpenGLES2(); // Example deletes OpenGL context here.
}
void SingleWindowAppView::Uninitialize()
{
}

void SingleWindowAppView::OnActivated(CoreApplicationView^ applicationView, IActivatedEventArgs^ args)
{
	CoreWindow::GetForCurrentThread()->Activate();
}
void SingleWindowAppView::OnVisibilityChanged(CoreWindow^ sender, VisibilityChangedEventArgs^ args)
{
	isWindowVisible=args->Visible;
}
void SingleWindowAppView::OnWindowClosed(CoreWindow^ sender, CoreWindowEventArgs^ args)
{
	isWindowClosed=true;
}

void SingleWindowAppView::InitializeOpenGLES2(void)
{
	// Cut & Pasted from OpenGL ES2 example.
	// Modified so that it selects the maximum depth bits.
	EGLint configAttributes[]=
	{
		EGL_RED_SIZE,8,
		EGL_GREEN_SIZE,8,
		EGL_BLUE_SIZE,8,
		EGL_ALPHA_SIZE,8,
		EGL_DEPTH_SIZE,24,
		EGL_STENCIL_SIZE,8,
		EGL_NONE
	};

	const EGLint contextAttributes[]=
	{
		EGL_CONTEXT_CLIENT_VERSION,2,
		EGL_NONE
	};

	const EGLint surfaceAttributes[]=
	{
		EGL_ANGLE_SURFACE_RENDER_TO_BACK_BUFFER,EGL_TRUE,
		EGL_NONE
	};

	EGLConfig config = NULL;

	// According to the example, the program should try to initialize OpenGL ES context with
	// EGL to D3D11 Feature Level 10_0+ first.  If it fails, try EGL to D3D11 Feature Level 9_3.
	// If it fails, try EGL to D3D11 Feature Level 11_0 on WARP.
	if(InitializeEGLDisplayDefault()!=EGL_TRUE &&
	   InitializeEGLDisplay(0)!=EGL_TRUE &&
	   InitializeEGLDisplay(1)!=EGL_TRUE &&
	   InitializeEGLDisplay(2)!=EGL_TRUE)
	{
		throw Exception::CreateException(E_FAIL, L"Failed to initialize EGL");
	}

	// The sample code given with the OpenGL ES template uses 8-bit Z-buffer, 
	// which is pretty much useless.
	// Ideallly want to have 64-bit (or as precise as coordinates, in many cases 32-bit), 
	// but here get whatever available.
	EGLint numConfigs = 0;
	for(int depthBit=32; 8<=depthBit; depthBit-=8)
	{
		SetUpConfigAttrib(configAttributes,depthBit);
		if(eglChooseConfig(mEglDisplay,configAttributes,&config,1,&numConfigs)==EGL_TRUE && 0!=numConfigs)
		{
			printf("%d Depth Bits\n",depthBit);
			break;
		}
	}
	if(0==numConfigs)
	{
		throw Exception::CreateException(E_FAIL,L"Failed to choose first EGLConfig");
	}

	PropertySet^ surfaceCreationProperties = ref new PropertySet();
	surfaceCreationProperties->Insert(ref new String(EGLNativeWindowTypeProperty), hWnd.Get());

	mEglSurface=eglCreateWindowSurface(mEglDisplay,config,reinterpret_cast<IInspectable*>(surfaceCreationProperties),surfaceAttributes);
	if(mEglSurface==EGL_NO_SURFACE)
	{
		throw Exception::CreateException(E_FAIL,L"Failed to create EGL fullscreen surface");
	}

	mEglContext=eglCreateContext(mEglDisplay,config,EGL_NO_CONTEXT,contextAttributes);
	if(mEglContext==EGL_NO_CONTEXT)
	{
		throw Exception::CreateException(E_FAIL,L"Failed to create EGL context");
	}

	if(eglMakeCurrent(mEglDisplay,mEglSurface,mEglSurface,mEglContext)==EGL_FALSE)
	{
		throw Exception::CreateException(E_FAIL,L"Failed to make fullscreen EGLSurface current");
	}
}

EGLBoolean SingleWindowAppView::InitializeEGLDisplayDefault(void)
{
	// The sample code tells that eglGetPlatformDisplayEXT is use in place for eglGetDisplay to pass display attributes.
	// But, why not trying eglGetDisplay first?
	mEglDisplay=eglGetDisplay(EGL_DEFAULT_DISPLAY);
	if(EGL_NO_DISPLAY!=mEglDisplay)
	{
		auto b=eglInitialize(mEglDisplay,NULL,NULL);
		printf("Looks like the default eglGetDisplay is good enough.  Isn't it?");
		return b;
	}
	return EGL_FALSE;
}

EGLBoolean SingleWindowAppView::InitializeEGLDisplay(int level)
{
	PFNEGLGETPLATFORMDISPLAYEXTPROC eglGetPlatformDisplayEXT=
		reinterpret_cast<PFNEGLGETPLATFORMDISPLAYEXTPROC>(eglGetProcAddress("eglGetPlatformDisplayEXT"));
	if (!eglGetPlatformDisplayEXT)
	{
		throw Exception::CreateException(E_FAIL, L"Failed to get function eglGetPlatformDisplayEXT");
	}

	auto attrib=MakeDisplayAttribute(level);

	mEglDisplay=eglGetPlatformDisplayEXT(EGL_PLATFORM_ANGLE_ANGLE,EGL_DEFAULT_DISPLAY,attrib.data());
	if(mEglDisplay!=EGL_NO_DISPLAY)
	{
		return eglInitialize(mEglDisplay,NULL,NULL);
	}
	return EGL_FALSE;
}

std::vector <EGLint> SingleWindowAppView::MakeDisplayAttribute(int level) const
{
	const EGLint commonDisplayAttributePre[]=
	{
		EGL_PLATFORM_ANGLE_TYPE_ANGLE,EGL_PLATFORM_ANGLE_TYPE_D3D11_ANGLE,
	};
	const EGLint commonDisplayAttributePost[]=
	{
		EGL_ANGLE_DISPLAY_ALLOW_RENDER_TO_BACK_BUFFER,EGL_TRUE,
		EGL_PLATFORM_ANGLE_ENABLE_AUTOMATIC_TRIM_ANGLE,EGL_TRUE,
		EGL_NONE,
	};

	std::vector <EGLint> attrib;
	for(auto v : commonDisplayAttributePre)
	{
		attrib.push_back(v);
	}
	switch(level)
	{
	case 0:
		break;
	case 1:  // FL9.  I don't know exactly what FL9 stands for, but as example says.
		attrib.push_back(EGL_PLATFORM_ANGLE_MAX_VERSION_MAJOR_ANGLE);
		attrib.push_back(9);
		attrib.push_back(EGL_PLATFORM_ANGLE_MAX_VERSION_MINOR_ANGLE);
		attrib.push_back(3);
		break;
	case 2:  // warp.  I don't know exactly what it means by "warp", but as example says.
		attrib.push_back(EGL_PLATFORM_ANGLE_DEVICE_TYPE_ANGLE);
		attrib.push_back(EGL_PLATFORM_ANGLE_DEVICE_TYPE_WARP_ANGLE);
		break;
	}
	for(auto v : commonDisplayAttributePost)
	{
		attrib.push_back(v);
	}
	return attrib;
}

void SingleWindowAppView::SetUpConfigAttrib(EGLint configAttrib[],int depthBit)
{
	for(int i=0; configAttrib[i]!=EGL_NONE; i+=2)
	{
		if(EGL_DEPTH_SIZE==configAttrib[i])
		{
			configAttrib[i+1]=depthBit;
		}
	}
}

void SingleWindowAppView::CleanUpOpenGLES2(void)
{
	// Cut & Pasted from OpenGL ES2 Example.
	if(mEglDisplay!=EGL_NO_DISPLAY && mEglSurface!=EGL_NO_SURFACE)
	{
		eglDestroySurface(mEglDisplay,mEglSurface);
		mEglSurface=EGL_NO_SURFACE;
	}
	if(mEglDisplay!=EGL_NO_DISPLAY && mEglContext!=EGL_NO_CONTEXT)
	{
		eglDestroyContext(mEglDisplay,mEglContext);
		mEglContext=EGL_NO_CONTEXT;
	}
	if(mEglDisplay!=EGL_NO_DISPLAY)
	{
		eglTerminate(mEglDisplay);
		mEglDisplay=EGL_NO_DISPLAY;
	}
}

ref class SingleWindowAppSource sealed : Windows::ApplicationModel::Core::IFrameworkViewSource
{
public:
	virtual Windows::ApplicationModel::Core::IFrameworkView^ CreateView()
	{
		return ref new SingleWindowAppView();
	}
};

[Platform::MTAThread]
int main(Platform::Array<Platform::String^>^)
{
	auto singleWindowAppSource = ref new SingleWindowAppSource();
	CoreApplication::Run(singleWindowAppSource);
	return 0;
}

