#include "uwpdirlist.h"
#include <thread>

#include <ppltasks.h> // To use Concurrency namespace

// https://social.msdn.microsoft.com/Forums/sqlserver/en-US/02299772-22a2-451a-9ae7-72dad6d3396c/ctaskwait-in-a-catchblock-throws-exception?forum=wpdevelop
// This discussion gave me a clue!  task<>::wait fails if in the GUI (main) thread.
// Then, all I need to do is start a new thread!

// http://stackoverflow.com/questions/19917466/how-to-wait-for-an-iasyncaction
// This solution, running an event-loop, also worked.  But, starting and waiting for a thread is
// cleaner and simpler.  Sorry users, please wait for a few second when I need a directory listing.

// Also, if the folder doesn't exist, it throws an exception instead of just returning a nullptr.
// (Give me a break!)  I am using try and catch inside the continuation.

using namespace Windows::Storage;
using namespace Windows::Foundation::Collections;  // To use IVectorView
using namespace Concurrency;                       // To use create_task

bool YsUwpFileList::MakeFileList(const wchar_t fulPath[])
{
	std::thread t(&YsUwpFileList::RunInThread,this,fulPath);
	t.join();
	return this->res;
}
void YsUwpFileList::RunInThread(const wchar_t fulPath[])
{
	// Note:
	//   Unlike Win32 functions, UWP functions do not accept '/' as separator.  It needs to be '\\'.
	//   Also, current directory '.' and parent directory '..' apparently is not recognized.
	//   Which only makes porting programs difficult.
	//   If Microsoft person is reading this, just fix it.  Take ".", "..", and '/'.
	//
	//   If you need to use relative path, _wgetcwd still seems to work.  So, assemble a full-path name
	//   and then give to this class.

	this->fileList.clear();
	this->res=true;
	Platform::String^ uwpPath=ref new Platform::String(fulPath);

	auto getFolderTask=create_task(Windows::Storage::StorageFolder::GetFolderFromPathAsync(uwpPath));
	getFolderTask.wait();
	try
	{
		auto folder=getFolderTask.get();
		auto getItemsTask=create_task(folder->GetItemsAsync());
		getItemsTask.wait();
		auto itemList=getItemsTask.get();

		for(int i=0; i<itemList->Size; ++i)
		{
			auto itm=itemList->GetAt(i);
			bool isDir=false;
			if((itm->Attributes&FileAttributes::Directory)==FileAttributes::Directory)
			{
				isDir=true;
			}
			std::basic_string <wchar_t> fn=itm->Name->Data();

			File newFile;
			fileList.push_back(newFile);
			fileList.back().nm=fn;
			fileList.back().isDir=isDir;
		}
	}
	catch(Platform::COMException^)
	{
		this->res=false;
	}
}
